"""Script to label GPU nodes in a Kubernetes cluster for use with SkyPilot"""
import argparse
import hashlib
import os
import subprocess
from typing import Dict, Optional, Tuple

import colorama
import yaml

import sky
from sky.adaptors import kubernetes
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.utils import rich_utils


def _format_string(str_to_format: str, colorama_format: str) -> str:
    return f'{colorama_format}{str_to_format}{colorama.Style.RESET_ALL}'


def cleanup(context: Optional[str] = None) -> Tuple[bool, str]:
    """Deletes all Kubernetes resources created by this script

    Used to provide idempotency when the script is run multiple times. Also
    invoked if --cleanup is passed to the script.
    """
    # Delete any existing GPU labeler Kubernetes resources:
    del_command = ('kubectl delete pods,services,deployments,jobs,daemonsets,'
                   'replicasets,configmaps,secrets,pv,pvc,clusterrole,'
                   'serviceaccount,clusterrolebinding -n kube-system '
                   '-l job=sky-gpu-labeler')
    if context:
        del_command += f' --context {context}'
    success = False
    reason = ''
    with rich_utils.client_status('Cleaning up existing GPU labeling '
                                  'resources'):
        try:
            subprocess.run(del_command.split(), check=True, capture_output=True)
            success = True
        except subprocess.CalledProcessError as e:
            output = e.output.decode('utf-8')
            reason = 'Error deleting existing GPU labeler resources: ' + output
        return success, reason


def get_node_hash(node_name: str):
    # Generates a 32 character md5 hash from a string
    md5_hash = hashlib.md5(node_name.encode()).hexdigest()
    return md5_hash[:32]


def label(context: Optional[str] = None, wait_for_completion: bool = True):
    deletion_success, reason = cleanup(context=context)
    if not deletion_success:
        print(reason)
        return

    unlabeled_gpu_nodes = kubernetes_utils.get_unlabeled_accelerator_nodes(
        context=context)

    if not unlabeled_gpu_nodes:
        print('No unlabeled GPU nodes found in the cluster. If you have '
              'unlabeled GPU nodes, please ensure that they have the resource '
              f'`{kubernetes_utils.get_gpu_resource_key(context)}: '
              '<number of GPUs>` in their capacity.')
        return

    print(
        _format_string(
            f'Found {len(unlabeled_gpu_nodes)} '
            'unlabeled GPU nodes in the cluster', colorama.Fore.YELLOW))

    sky_dir = os.path.dirname(sky.__file__)
    manifest_dir = os.path.join(sky_dir, 'utils/kubernetes')

    # Apply the RBAC manifest using kubectl since it contains multiple resources
    with rich_utils.client_status('Setting up GPU labeling'):
        rbac_manifest_path = os.path.join(manifest_dir,
                                          'k8s_gpu_labeler_setup.yaml')
        try:
            apply_command = ['kubectl', 'apply', '-f', rbac_manifest_path]
            if context:
                apply_command += ['--context', context]
            subprocess.check_output(apply_command)
        except subprocess.CalledProcessError as e:
            output = e.output.decode('utf-8')
            print('Error setting up GPU labeling: ' + output)
            return

    jobs_to_node_names: Dict[str, str] = {}
    with rich_utils.client_status('Creating GPU labeler jobs'):
        batch_v1 = kubernetes.batch_api(context=context)
        # Load the job manifest
        job_manifest_path = os.path.join(manifest_dir,
                                         'k8s_gpu_labeler_job.yaml')

        with open(job_manifest_path, 'r', encoding='utf-8') as file:
            job_manifest = yaml.safe_load(file)

        # Check if the 'nvidia' RuntimeClass exists
        try:
            nvidia_exists = kubernetes_utils.check_nvidia_runtime_class(
                context=context)
        except Exception as e:  # pylint: disable=broad-except
            print('Error occurred while checking for nvidia RuntimeClass: '
                  f'{str(e)}')
            print('Continuing without using nvidia RuntimeClass. '
                  'This may fail on K3s clusters. '
                  'For more details, refer to K3s deployment notes at: '
                  'https://docs.skypilot.co/en/latest/reference/kubernetes/kubernetes-setup.html')  # pylint: disable=line-too-long
            nvidia_exists = False

        if nvidia_exists:
            print('Using nvidia RuntimeClass for GPU labeling.')
            job_manifest['spec']['template']['spec'][
                'runtimeClassName'] = 'nvidia'
        else:
            print('Using default RuntimeClass for GPU labeling.')

        for node in unlabeled_gpu_nodes:
            node_name = node.metadata.name

            # Modify the job manifest for the current node
            job_name = ('sky-gpu-labeler-'
                        f'{get_node_hash(node_name)}')
            jobs_to_node_names[job_name] = node_name
            job_manifest['metadata']['name'] = job_name

            job_manifest['spec']['template']['spec']['nodeSelector'] = {
                'kubernetes.io/hostname': node_name
            }
            namespace = job_manifest['metadata']['namespace']

            # Create the job for this node`
            batch_v1.create_namespaced_job(namespace, job_manifest)
            print(
                _format_string(f'Created GPU labeler job for node {node_name}',
                               colorama.Style.DIM))

    context_str = f' --context {context}' if context else ''

    if wait_for_completion:
        # Wait for the job to complete
        with rich_utils.client_status(
                'Waiting for GPU labeler jobs to complete'):
            success = wait_for_jobs_completion(jobs_to_node_names,
                                               'kube-system',
                                               context=context)
        if success:
            print(
                _format_string('✅ GPU labeling completed successfully',
                               colorama.Fore.GREEN))
        else:
            print(_format_string('❌ GPU labeling failed', colorama.Fore.RED))
        cleanup(context=context)
    else:
        print(
            f'GPU labeling started - this may take 10 min or more to complete.'
            '\nTo check the status of GPU labeling jobs, run '
            f'`kubectl get jobs -n kube-system '
            f'-l job=sky-gpu-labeler{context_str}`'
            '\nYou can check if nodes have been labeled by running '
            f'`kubectl describe nodes{context_str}` '
            'and looking for labels of the format '
            '`skypilot.co/accelerator: <gpu_name>`. ')


def wait_for_jobs_completion(jobs_to_node_names: Dict[str, str],
                             namespace: str,
                             context: Optional[str] = None,
                             timeout: int = 60 * 20):
    """Waits for a Kubernetes Job to complete or fail.

    Args:
        jobs_to_node_names: A dictionary mapping job names to node names.
        namespace: The namespace the Job is in (default: "default").
        timeout: Timeout in seconds (default: 1200 seconds = 20 minutes).

    Returns:
        True if the Job completed successfully, False if it failed or timed out.
    """
    batch_v1 = kubernetes.batch_api(context=context)
    w = kubernetes.watch()
    completed_jobs = []
    for event in w.stream(func=batch_v1.list_namespaced_job,
                          namespace=namespace,
                          timeout_seconds=timeout):
        job = event['object']
        job_name = job.metadata.name
        if job_name in jobs_to_node_names:
            node_name = jobs_to_node_names[job_name]
            if job.status and job.status.completion_time:
                print(
                    _format_string(
                        f'GPU labeler job for node {node_name} '
                        'completed successfully', colorama.Style.DIM))
                completed_jobs.append(job_name)
                num_remaining_jobs = len(jobs_to_node_names) - len(
                    completed_jobs)
                if num_remaining_jobs == 0:
                    w.stop()
                    return True
            elif job.status and job.status.failed:
                print(
                    _format_string(
                        f'GPU labeler job for node {node_name} failed',
                        colorama.Style.DIM))
                w.stop()
                return False
    print(
        _format_string(
            f'Timed out after waiting {timeout} seconds '
            'for job to complete', colorama.Style.DIM))
    return False  #Timed out


def main():
    parser = argparse.ArgumentParser(
        description='Labels GPU nodes in a Kubernetes cluster for use with '
        'SkyPilot. Operates by running a job on each node that '
        'parses nvidia-smi and patches the node with new labels. '
        'Labels created are of the format '
        'skypilot.co/accelerator: <gpu_name>. Automatically '
        'creates a service account and cluster role binding with '
        'permissions to list nodes and create labels.')
    parser.add_argument('--cleanup',
                        action='store_true',
                        help='delete all GPU labeler resources in the '
                        'Kubernetes cluster.')
    parser.add_argument('--context',
                        type=str,
                        help='the context to use for the Kubernetes cluster.')
    parser.add_argument('--async',
                        dest='async_completion',
                        action='store_true',
                        help='do not wait for the GPU labeling to complete.')
    args = parser.parse_args()
    context = None
    if args.context:
        context = args.context

    # Check if kubectl is installed and kubeconfig is set up
    prereq_ok, reason = kubernetes_utils.check_credentials(context=context)
    if not prereq_ok:
        print(reason)
        return

    if args.cleanup:
        cleanup(context=context)
    else:
        label(context=context, wait_for_completion=not args.async_completion)


if __name__ == '__main__':
    main()



"""Volume utils."""
import abc
from datetime import datetime
from typing import Any, Dict, List, Optional

import prettytable

from sky import sky_logging
from sky.skylet import constants
from sky.utils import common_utils
from sky.utils import log_utils
from sky.utils import volume

logger = sky_logging.init_logger(__name__)


def _get_infra_str(cloud: Optional[str], region: Optional[str],
                   zone: Optional[str]) -> str:
    """Get the infrastructure string for the volume."""
    infra = ''
    if cloud:
        infra += cloud
    if region:
        infra += f'/{region}'
    if zone:
        infra += f'/{zone}'
    return infra


class VolumeTable(abc.ABC):
    """The volume table."""

    @abc.abstractmethod
    def format(self) -> str:
        """Format the volume table for display."""
        pass


class PVCVolumeTable(VolumeTable):
    """The PVC volume table."""

    def __init__(self, volumes: List[Dict[str, Any]], show_all: bool = False):
        super().__init__()
        self.table = self._create_table(show_all)
        self._add_rows(volumes, show_all)

    def _create_table(self, show_all: bool = False) -> prettytable.PrettyTable:
        """Create the PVC volume table."""
        #  If show_all is False, show the table with the columns:
        #   NAME, TYPE, INFRA, SIZE, USER, WORKSPACE,
        #   AGE, STATUS, LAST_USE, USED_BY
        #  If show_all is True, show the table with the columns:
        #   NAME, TYPE, INFRA, SIZE, USER, WORKSPACE,
        #   AGE, STATUS, LAST_USE, USED_BY, NAME_ON_CLOUD
        #   STORAGE_CLASS, ACCESS_MODE

        if show_all:
            columns = [
                'NAME',
                'TYPE',
                'INFRA',
                'SIZE',
                'USER',
                'WORKSPACE',
                'AGE',
                'STATUS',
                'LAST_USE',
                'USED_BY',
                'NAME_ON_CLOUD',
                'STORAGE_CLASS',
                'ACCESS_MODE',
            ]
        else:
            columns = [
                'NAME',
                'TYPE',
                'INFRA',
                'SIZE',
                'USER',
                'WORKSPACE',
                'AGE',
                'STATUS',
                'LAST_USE',
                'USED_BY',
            ]

        table = log_utils.create_table(columns)
        return table

    def _add_rows(self,
                  volumes: List[Dict[str, Any]],
                  show_all: bool = False) -> None:
        """Add rows to the PVC volume table."""
        for row in volumes:
            # Convert last_attached_at timestamp to human readable string
            last_attached_at = row.get('last_attached_at')
            if last_attached_at is not None:
                last_attached_at_str = datetime.fromtimestamp(
                    last_attached_at).strftime('%Y-%m-%d %H:%M:%S')
            else:
                last_attached_at_str = '-'
            size = row.get('size', '')
            if size:
                size = f'{size}Gi'
            usedby_str = '-'
            usedby_clusters = row.get('usedby_clusters')
            usedby_pods = row.get('usedby_pods')
            if usedby_clusters:
                usedby_str = f'{", ".join(usedby_clusters)}'
            elif usedby_pods:
                usedby_str = f'{", ".join(usedby_pods)}'
            if show_all:
                usedby = usedby_str
            else:
                usedby = common_utils.truncate_long_string(
                    usedby_str, constants.USED_BY_TRUNC_LENGTH)
            infra = _get_infra_str(row.get('cloud'), row.get('region'),
                                   row.get('zone'))
            table_row = [
                row.get('name', ''),
                row.get('type', ''),
                infra,
                size,
                row.get('user_name', '-'),
                row.get('workspace', '-'),
                log_utils.human_duration(row.get('launched_at', 0)),
                row.get('status', ''),
                last_attached_at_str,
                usedby,
            ]
            if show_all:
                table_row.append(row.get('name_on_cloud', ''))
                table_row.append(
                    row.get('config', {}).get('storage_class_name', '-'))
                table_row.append(row.get('config', {}).get('access_mode', ''))

            self.table.add_row(table_row)

    def format(self) -> str:
        """Format the PVC volume table for display."""
        return str(self.table)


def format_volume_table(volumes: List[Dict[str, Any]],
                        show_all: bool = False) -> str:
    """Format the volume table for display.

    Args:
        volume_table (dict): The volume table.

    Returns:
        str: The formatted volume table.
    """
    volumes_per_type: Dict[str, List[Dict[str, Any]]] = {}
    supported_volume_types = [
        volume_type.value for volume_type in volume.VolumeType
    ]
    for row in volumes:
        volume_type = row.get('type', '')
        if volume_type in supported_volume_types:
            if volume_type not in volumes_per_type:
                volumes_per_type[volume_type] = []
            volumes_per_type[volume_type].append(row)
        else:
            logger.warning(f'Unknown volume type: {volume_type}')
            continue
    table_str = ''
    for volume_type, volume_list in volumes_per_type.items():
        if volume_type == volume.VolumeType.PVC.value:
            table = PVCVolumeTable(volume_list, show_all)
            table_str += table.format()
    if table_str:
        return table_str
    else:
        return 'No existing volumes.'



""" Nebius Cloud. """
import json
import os
import typing
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from sky import catalog
from sky import clouds
from sky import exceptions
from sky import skypilot_config
from sky.adaptors import nebius
from sky.provision.nebius import constants as nebius_constants
from sky.utils import annotations
from sky.utils import registry
from sky.utils import resources_utils

if typing.TYPE_CHECKING:
    from sky import resources as resources_lib
    from sky.utils import volume as volume_lib

_INDENT_PREFIX = '    '


def nebius_profile_in_aws_cred_and_config() -> bool:
    """Checks if Nebius Object Storage profile is set in aws credentials
    and profile."""

    credentials_path = os.path.expanduser('~/.aws/credentials')
    nebius_profile_exists_in_credentials = False
    if os.path.isfile(credentials_path):
        with open(credentials_path, 'r', encoding='utf-8') as file:
            for line in file:
                if f'[{nebius.NEBIUS_PROFILE_NAME}]' in line:
                    nebius_profile_exists_in_credentials = True

    config_path = os.path.expanduser('~/.aws/config')
    nebius_profile_exists_in_config = False
    if os.path.isfile(config_path):
        with open(config_path, 'r', encoding='utf-8') as file:
            for line in file:
                if f'[profile {nebius.NEBIUS_PROFILE_NAME}]' in line:
                    nebius_profile_exists_in_config = True

    return (nebius_profile_exists_in_credentials and
            nebius_profile_exists_in_config)


@registry.CLOUD_REGISTRY.register
class Nebius(clouds.Cloud):
    """Nebius GPU Cloud"""
    _REPR = 'Nebius'
    _CLOUD_UNSUPPORTED_FEATURES = {
        clouds.CloudImplementationFeatures.AUTODOWN:
            ('Autodown not supported. Can\'t delete OS disk.'),
        clouds.CloudImplementationFeatures.CLONE_DISK_FROM_CLUSTER:
            (f'Migrating disk is currently not supported on {_REPR}.'),
        clouds.CloudImplementationFeatures.CUSTOM_DISK_TIER:
            (f'Custom disk tier is currently not supported on {_REPR}.'),
        clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER:
            ('Custom network tier is currently not supported on Nebius.'),
        clouds.CloudImplementationFeatures.HIGH_AVAILABILITY_CONTROLLERS:
            ('High availability controllers are not supported on Nebius.'),
        clouds.CloudImplementationFeatures.CUSTOM_MULTI_NETWORK:
            ('Customized multiple network interfaces are not supported on '
             f'{_REPR}.'),
    }
    # Nebius maximum instance name length defined as <= 63 as a hostname length
    # 63 - 8 - 5 = 50 characters since
    # we add 4 character from UUID to make uniq `-xxxx`
    # our provisioner adds additional `-worker`.
    _MAX_CLUSTER_NAME_LEN_LIMIT = 50
    _regions: List[clouds.Region] = []

    # Using the latest SkyPilot provisioner API to provision and check status.
    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT

    @classmethod
    def _unsupported_features_for_resources(
        cls, resources: 'resources_lib.Resources'
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        unsupported = cls._CLOUD_UNSUPPORTED_FEATURES.copy()

        # Check if the accelerators support InfiniBand (H100 or H200) and 8 GPUs
        if resources.accelerators is not None:
            for acc_name, acc_count in resources.accelerators.items():
                if acc_name.lower() in ('h100', 'h200') and acc_count == 8:
                    # Remove CUSTOM_NETWORK_TIER from unsupported features for
                    # InfiniBand-capable accelerators
                    unsupported.pop(
                        clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER,
                        None)
                    break

        return unsupported

    @classmethod
    def _max_cluster_name_length(cls) -> Optional[int]:
        return cls._MAX_CLUSTER_NAME_LEN_LIMIT

    @classmethod
    def regions_with_offering(cls, instance_type: str,
                              accelerators: Optional[Dict[str, int]],
                              use_spot: bool, region: Optional[str],
                              zone: Optional[str]) -> List[clouds.Region]:
        assert zone is None, 'Nebius does not support zones.'
        del accelerators, zone  # unused
        regions = catalog.get_region_zones_for_instance_type(
            instance_type, use_spot, 'nebius')

        if region is not None:
            regions = [r for r in regions if r.name == region]
        return regions

    @classmethod
    def get_vcpus_mem_from_instance_type(
        cls,
        instance_type: str,
    ) -> Tuple[Optional[float], Optional[float]]:
        return catalog.get_vcpus_mem_from_instance_type(instance_type,
                                                        clouds='nebius')

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[None]:
        del num_nodes  # unused
        regions = cls.regions_with_offering(instance_type,
                                            accelerators,
                                            use_spot,
                                            region=region,
                                            zone=None)
        for r in regions:
            assert r.zones is None, r
            yield r.zones

    def instance_type_to_hourly_cost(self,
                                     instance_type: str,
                                     use_spot: bool,
                                     region: Optional[str] = None,
                                     zone: Optional[str] = None) -> float:
        return catalog.get_hourly_cost(instance_type,
                                       use_spot=use_spot,
                                       region=region,
                                       zone=zone,
                                       clouds='nebius')

    def accelerators_to_hourly_cost(self,
                                    accelerators: Dict[str, int],
                                    use_spot: bool,
                                    region: Optional[str] = None,
                                    zone: Optional[str] = None) -> float:
        """Returns the hourly cost of the accelerators, in dollars/hour."""
        del accelerators, use_spot, region, zone  # unused
        return 0.0

    def get_egress_cost(self, num_gigabytes: float) -> float:
        return 0.0

    def __repr__(self):
        return self._REPR

    def is_same_cloud(self, other: clouds.Cloud) -> bool:
        # Returns true if the two clouds are the same cloud type.
        return isinstance(other, Nebius)

    @classmethod
    def get_default_instance_type(cls,
                                  cpus: Optional[str] = None,
                                  memory: Optional[str] = None,
                                  disk_tier: Optional[
                                      resources_utils.DiskTier] = None,
                                  region: Optional[str] = None,
                                  zone: Optional[str] = None) -> Optional[str]:
        """Returns the default instance type for Nebius."""
        return catalog.get_default_instance_type(cpus=cpus,
                                                 memory=memory,
                                                 disk_tier=disk_tier,
                                                 region=region,
                                                 zone=zone,
                                                 clouds='nebius')

    @classmethod
    def get_accelerators_from_instance_type(
        cls,
        instance_type: str,
    ) -> Optional[Dict[str, Union[int, float]]]:
        return catalog.get_accelerators_from_instance_type(instance_type,
                                                           clouds='nebius')

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        return None

    def make_deploy_resources_variables(
        self,
        resources: 'resources_lib.Resources',
        cluster_name: resources_utils.ClusterName,
        region: 'clouds.Region',
        zones: Optional[List['clouds.Zone']],
        num_nodes: int,
        dryrun: bool = False,
        volume_mounts: Optional[List['volume_lib.VolumeMount']] = None,
    ) -> Dict[str, Any]:
        del dryrun, cluster_name
        assert zones is None, ('Nebius does not support zones', zones)

        resources = resources.assert_launchable()
        acc_dict = self.get_accelerators_from_instance_type(
            resources.instance_type)
        custom_resources = resources_utils.make_ray_custom_resources_str(
            acc_dict)
        platform, _ = resources.instance_type.split('_')

        # Selecting image_family by platform
        # https://docs.nebius.com/compute/storage/boot-disk-images
        if platform.startswith('cpu'):
            image_family = 'ubuntu24.04-driverless'
        elif platform.startswith('gpu'):
            image_family = 'ubuntu24.04-cuda12'
        else:
            raise RuntimeError('Unsupported instance type for Nebius cloud:'
                               f' {resources.instance_type}')

        config_fs = skypilot_config.get_effective_region_config(
            cloud='nebius',
            region=region.name,
            keys=('filesystems',),
            default_value=[])
        resources_vars_fs = []
        for i, fs in enumerate(config_fs):
            resources_vars_fs.append({
                'filesystem_id': fs['filesystem_id'],
                'filesystem_attach_mode': fs.get('attach_mode', 'READ_WRITE'),
                'filesystem_mount_path': fs.get(
                    'mount_path', f'/mnt/filesystem-skypilot-{i+1}'),
                'filesystem_mount_tag': f'filesystem-skypilot-{i+1}'
            })

        resources_vars: Dict[str, Any] = {
            'instance_type': resources.instance_type,
            'custom_resources': custom_resources,
            'region': region.name,
            'image_id': image_family,
            # Nebius does not support specific zones.
            'zones': None,
            'use_spot': resources.use_spot,
            'filesystems': resources_vars_fs,
            'network_tier': resources.network_tier
        }

        docker_run_options = []

        if acc_dict is not None:
            # Nebius cloud's docker runtime information does not contain
            # 'nvidia-container-runtime', causing no GPU option to be added to
            # the docker run command. We patch this by adding it here.
            docker_run_options.append('--gpus all')

            # Check for InfiniBand support with network_tier: best
            is_infiniband_capable = (
                platform in nebius_constants.INFINIBAND_INSTANCE_PLATFORMS)
            if (is_infiniband_capable and
                    resources.network_tier == resources_utils.NetworkTier.BEST):
                # For Docker containers, add InfiniBand device access and
                # IPC_LOCK capability
                if resources.extract_docker_image() is not None:
                    docker_run_options.extend(
                        nebius_constants.INFINIBAND_DOCKER_OPTIONS)

                    # Add InfiniBand environment variables to docker run options
                    for env_var, env_value in (
                            nebius_constants.INFINIBAND_ENV_VARS.items()):
                        docker_run_options.extend(
                            ['-e', f'{env_var}={env_value}'])

                # For all InfiniBand-capable instances, add env variables
                resources_vars[
                    'env_vars'] = nebius_constants.INFINIBAND_ENV_VARS

        if docker_run_options:
            resources_vars['docker_run_options'] = docker_run_options

        return resources_vars

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> 'resources_utils.FeasibleResources':
        """Returns a list of feasible resources for the given resources."""
        if resources.instance_type is not None:
            assert resources.is_launchable(), resources
            resources = resources.copy(accelerators=None)
            return resources_utils.FeasibleResources([resources], [], None)

        def _make(instance_list):
            resource_list = []
            for instance_type in instance_list:
                r = resources.copy(
                    cloud=Nebius(),
                    instance_type=instance_type,
                    accelerators=None,
                    cpus=None,
                )
                resource_list.append(r)
            return resource_list

        # Currently, handle a filter on accelerators only.
        accelerators = resources.accelerators
        if accelerators is None:
            # Return a default instance type
            default_instance_type = Nebius.get_default_instance_type(
                cpus=resources.cpus,
                memory=resources.memory,
                disk_tier=resources.disk_tier,
                region=resources.region,
                zone=resources.zone)
            if default_instance_type is None:
                # TODO: Add hints to all return values in this method to help
                #  users understand why the resources are not launchable.
                return resources_utils.FeasibleResources([], [], None)
            else:
                return resources_utils.FeasibleResources(
                    _make([default_instance_type]), [], None)

        assert len(accelerators) == 1, resources
        acc, acc_count = list(accelerators.items())[0]
        (instance_list,
         fuzzy_candidate_list) = catalog.get_instance_type_for_accelerator(
             acc,
             acc_count,
             use_spot=resources.use_spot,
             cpus=resources.cpus,
             region=resources.region,
             zone=resources.zone,
             clouds='nebius')
        if instance_list is None:
            return resources_utils.FeasibleResources([], fuzzy_candidate_list,
                                                     None)
        return resources_utils.FeasibleResources(_make(instance_list),
                                                 fuzzy_candidate_list, None)

    @classmethod
    def _check_compute_credentials(
            cls) -> Tuple[bool, Optional[Union[str, Dict[str, str]]]]:
        """Checks if the user has access credentials to
        Nebius's compute service."""
        token_cred_msg = (
            f'{_INDENT_PREFIX}Credentials can be set up by running: \n'
            f'{_INDENT_PREFIX}  $ nebius iam get-access-token > {nebius.iam_token_path()} \n'  # pylint: disable=line-too-long
            f'{_INDENT_PREFIX} or generate  {nebius.credentials_path()} \n')

        tenant_msg = (f'{_INDENT_PREFIX} Copy your tenant ID from the web console and save it to file \n'  # pylint: disable=line-too-long
                      f'{_INDENT_PREFIX}  $ echo $NEBIUS_TENANT_ID_PATH > {nebius.tenant_id_path()} \n'  # pylint: disable=line-too-long
                      f'{_INDENT_PREFIX} Or if you have 1 tenant you can run:\n'  # pylint: disable=line-too-long
                      f'{_INDENT_PREFIX}  $ nebius --format json iam whoami|jq -r \'.user_profile.tenants[0].tenant_id\' > {nebius.tenant_id_path()} \n')  # pylint: disable=line-too-long
        if not nebius.is_token_or_cred_file_exist():
            return False, f'{token_cred_msg}'
        sdk = nebius.sdk()
        tenant_id = nebius.get_tenant_id()
        if tenant_id is None:
            return False, f'{tenant_msg}'
        try:
            service = nebius.iam().ProjectServiceClient(sdk)
            service.list(
                nebius.iam().ListProjectsRequest(parent_id=tenant_id)).wait()
        except nebius.request_error() as e:
            return False, (
                f'{e.status} \n'  # First line is indented by 4 spaces
                f'{token_cred_msg} \n'
                f'{tenant_msg}')
        return True, None

    @classmethod
    @annotations.lru_cache(scope='request')
    def _check_storage_credentials(
            cls) -> Tuple[bool, Optional[Union[str, Dict[str, str]]]]:
        """Checks if the user has access credentials to Nebius Object Storage.

        Returns:
            A tuple of a boolean value and a hint message where the bool
            is True when credentials needed for Nebius Object Storage is set.
            It is False when either of those are not set, which would hint
            with a string on unset credential.
        """
        hints = None
        if not nebius_profile_in_aws_cred_and_config():
            hints = (f'[{nebius.NEBIUS_PROFILE_NAME}] profile '
                     'is not set in ~/.aws/credentials.')
        if hints:
            hints += ' Run the following commands:'
            if not nebius_profile_in_aws_cred_and_config():
                hints += (
                    f'\n{_INDENT_PREFIX}  $ pip install boto3'
                    f'\n{_INDENT_PREFIX}  $ aws configure --profile nebius')
            hints += (
                f'\n{_INDENT_PREFIX}For more info: '
                'https://docs.skypilot.co/en/latest/getting-started/installation.html#nebius'  # pylint: disable=line-too-long
            )
        return (False, hints) if hints else (True, hints)

    def get_credential_file_mounts(self) -> Dict[str, str]:
        credential_file_mounts = {
            filepath: filepath
            for filepath in nebius.get_credential_file_paths()
        }
        if nebius_profile_in_aws_cred_and_config():
            credential_file_mounts['~/.aws/credentials'] = '~/.aws/credentials'
            credential_file_mounts['~/.aws/config'] = '~/.aws/config'
        return credential_file_mounts

    @classmethod
    def get_current_user_identity(cls) -> Optional[List[str]]:
        # NOTE: used for very advanced SkyPilot functionality
        # Can implement later if desired
        return None

    def instance_type_exists(self, instance_type: str) -> bool:
        return catalog.instance_type_exists(instance_type, 'nebius')

    def validate_region_zone(self, region: Optional[str], zone: Optional[str]):
        return catalog.validate_region_zone(region, zone, clouds='nebius')

    @classmethod
    def get_user_identities(cls) -> Optional[List[List[str]]]:
        """Returns the email address + project id of the active user."""
        nebius_workspace_config = json.dumps(
            skypilot_config.get_workspace_cloud('nebius'), sort_keys=True)
        return cls._get_user_identities(nebius_workspace_config)

    @classmethod
    @annotations.lru_cache(scope='request', maxsize=5)
    def _get_user_identities(
            cls, workspace_config: Optional[str]) -> Optional[List[List[str]]]:
        # We add workspace_config in args to avoid caching the identity for when
        # different workspace configs are used.
        del workspace_config  # Unused
        sdk = nebius.sdk()
        profile_client = nebius.iam().ProfileServiceClient(sdk)
        profile = nebius.sync_call(
            profile_client.get(nebius.iam().GetProfileRequest(),
                               timeout=nebius.READ_TIMEOUT))
        if profile.user_profile is not None:
            if profile.user_profile.attributes is None:
                raise exceptions.CloudUserIdentityError(
                    'Nebius profile is a UserProfile, but has no attributes: '
                    f'{profile.user_profile}')
            if profile.user_profile.attributes.email is None:
                raise exceptions.CloudUserIdentityError(
                    'Nebius profile is a UserProfile, but has no email: '
                    f'{profile.user_profile}')
            return [[profile.user_profile.attributes.email]]
        if profile.service_account_profile is not None:
            if profile.service_account_profile.info is None:
                raise exceptions.CloudUserIdentityError(
                    'Nebius profile is a ServiceAccountProfile, but has no '
                    f'info: {profile.service_account_profile}')
            if profile.service_account_profile.info.metadata is None:
                raise exceptions.CloudUserIdentityError(
                    'Nebius profile is a ServiceAccountProfile, but has no '
                    f'metadata: {profile.service_account_profile}')
            if profile.service_account_profile.info.metadata.name is None:
                raise exceptions.CloudUserIdentityError(
                    'Nebius profile is a ServiceAccountProfile, but has no '
                    f'name: {profile.service_account_profile}')
            return [[profile.service_account_profile.info.metadata.name]]
        if profile.anonymous_profile is not None:
            return None
        unknown_profile_type = profile.which_field_in_oneof('profile')
        raise exceptions.CloudUserIdentityError(
            f'Nebius profile is of an unknown type - {unknown_profile_type}')



"""Common utilities for the client."""

import contextlib
import dataclasses
import json
import logging
import math
import os
import pathlib
import tempfile
import time
import typing
from typing import Dict, Generator, Iterable
import uuid
import zipfile

from sky import sky_logging
from sky.adaptors import common as adaptors_common
from sky.client import service_account_auth
from sky.data import data_utils
from sky.data import storage_utils
from sky.server import common as server_common
from sky.server.requests import payloads
from sky.skylet import constants
from sky.utils import common_utils
from sky.utils import rich_utils
from sky.utils import subprocess_utils
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    import httpx
    import requests

    import sky
    from sky import dag as dag_lib
else:
    httpx = adaptors_common.LazyImport('httpx')
    requests = adaptors_common.LazyImport('requests')

logger = sky_logging.init_logger(__name__)

# The chunk size for downloading the logs from the API server.
_DOWNLOAD_CHUNK_BYTES = 8192
# The chunk size for the zip file to be uploaded to the API server. We split
# the zip file into chunks to avoid network issues for large request body that
# can be caused by NGINX's client_max_body_size.
_UPLOAD_CHUNK_BYTES = 512 * 1024 * 1024

FILE_UPLOAD_LOGS_DIR = os.path.join(constants.SKY_LOGS_DIRECTORY,
                                    'file_uploads')

# Connection timeout when sending requests to the API server.
API_SERVER_REQUEST_CONNECTION_TIMEOUT_SECONDS = 5


def download_logs_from_api_server(
        paths_on_api_server: Iterable[str],
        remote_machine_prefix: str = str(
            server_common.api_server_user_logs_dir_prefix()),
        local_machine_prefix: str = constants.SKY_LOGS_DIRECTORY
) -> Dict[str, str]:
    """Downloads the logs from the API server.

    Args:
        paths_on_api_server: The paths on the API server to download.
        remote_machine_prefix: The prefix of the remote machine to save the
        logs.
        local_machine_prefix: The prefix of the local machine to save the logs.

    Returns:
        A dictionary mapping the remote path on API server to the local path.
    """
    remote2local_path_dict = {
        remote_path: remote_path.replace(
            # TODO(zhwu): handling the replacement locally is not stable, and
            # may cause issues when we change the pattern of the remote path.
            # This should be moved to remote API server. A proper way might be
            # set the returned path to be started with a special prefix, instead
            # of using the `api_server_user_logs_dir_prefix()`.
            remote_machine_prefix,
            local_machine_prefix) for remote_path in paths_on_api_server
    }
    body = payloads.DownloadBody(folder_paths=list(paths_on_api_server),)
    response = requests.post(f'{server_common.get_server_url()}/download',
                             json=json.loads(body.model_dump_json()),
                             stream=True,
                             cookies=server_common.get_api_cookie_jar())
    if response.status_code == 200:
        remote_home_path = response.headers.get('X-Home-Path')
        assert remote_home_path is not None, response.headers
        with tempfile.NamedTemporaryFile(prefix='skypilot-logs-download-',
                                         delete=True) as temp_file:
            # Download the zip file from the API server to the local machine.
            for chunk in response.iter_content(
                    chunk_size=_DOWNLOAD_CHUNK_BYTES):
                temp_file.write(chunk)
            temp_file.flush()

            # Unzip the downloaded file and save the logs to the correct local
            # directory.
            with zipfile.ZipFile(temp_file, 'r') as zipf:
                for member in zipf.namelist():
                    # Determine the new path
                    zipped_filename = os.path.basename(member)
                    zipped_dir = os.path.dirname('/' + member)
                    local_dir = zipped_dir.replace(remote_home_path, '~')
                    for remote_path, local_path in remote2local_path_dict.items(
                    ):
                        if local_dir.startswith(remote_path):
                            local_dir = local_dir.replace(
                                remote_path, local_path)
                            break
                    else:
                        raise ValueError(f'Invalid folder path: {zipped_dir}')
                    new_path = pathlib.Path(
                        local_dir).expanduser().resolve() / zipped_filename
                    new_path.parent.mkdir(parents=True, exist_ok=True)
                    if member.endswith('/'):
                        # If it is a directory, we need to create it.
                        new_path.mkdir(parents=True, exist_ok=True)
                    else:
                        with zipf.open(member) as member_file:
                            new_path.write_bytes(member_file.read())

        return remote2local_path_dict
    else:
        raise Exception(
            f'Failed to download logs: {response.status_code} {response.text}')


# === Upload files to API server ===


class FileChunkIterator:
    """A file-like object that reads from a file in chunks."""

    def __init__(self, file_obj, chunk_size: int, chunk_index: int):
        self.file_obj = file_obj
        self.chunk_size = chunk_size
        self.chunk_index = chunk_index
        self.bytes_read = 0

    def __iter__(self):
        # Seek to the correct position for this chunk
        self.file_obj.seek(self.chunk_index * self.chunk_size)
        while self.bytes_read < self.chunk_size:
            # Read a smaller buffer size to keep memory usage low
            buffer_size = min(64 * 1024,
                              self.chunk_size - self.bytes_read)  # 64KB buffer
            data = self.file_obj.read(buffer_size)
            if not data:
                break
            self.bytes_read += len(data)
            yield data


@dataclasses.dataclass
class UploadChunkParams:
    client: 'httpx.Client'
    upload_id: str
    chunk_index: int
    total_chunks: int
    file_path: str
    upload_logger: logging.Logger
    log_file: str


def _upload_chunk_with_retry(params: UploadChunkParams) -> None:
    """Uploads a chunk of a zip file to the API server."""
    upload_logger = params.upload_logger
    upload_logger.info(
        f'Uploading chunk: {params.chunk_index + 1} / {params.total_chunks}')

    server_url = server_common.get_server_url()
    max_attempts = 3
    sa_headers = service_account_auth.get_service_account_headers()
    with open(params.file_path, 'rb') as f:
        for attempt in range(max_attempts):
            response = params.client.post(
                f'{server_url}/upload',
                params={
                    'user_hash': common_utils.get_user_hash(),
                    'upload_id': params.upload_id,
                    'chunk_index': str(params.chunk_index),
                    'total_chunks': str(params.total_chunks),
                },
                content=FileChunkIterator(f, _UPLOAD_CHUNK_BYTES,
                                          params.chunk_index),
                headers={
                    'Content-Type': 'application/octet-stream',
                    **sa_headers,
                },
                cookies=server_common.get_api_cookie_jar())
            if response.status_code == 200:
                data = response.json()
                status = data.get('status')
                msg = ('Uploaded chunk: '
                       f'{params.chunk_index + 1} / {params.total_chunks}')
                if status == 'uploading':
                    missing_chunks = data.get('missing_chunks')
                    if missing_chunks:
                        msg += f' - Waiting for chunks: {missing_chunks}'
                upload_logger.info(msg)
                return
            elif attempt < max_attempts - 1:
                upload_logger.error(
                    f'Failed to upload chunk: '
                    f'{params.chunk_index + 1} / {params.total_chunks}: '
                    f'{response.content.decode("utf-8")}')
                upload_logger.info(
                    f'Retrying... ({attempt + 1} / {max_attempts})')
                time.sleep(1)
            else:
                try:
                    response_details = response.json().get('detail')
                except Exception:  # pylint: disable=broad-except
                    response_details = response.content
                error_msg = (
                    f'Failed to upload chunk: {params.chunk_index + 1} / '
                    f'{params.total_chunks}: {response_details} '
                    f'(Status code: {response.status_code})')
                upload_logger.error(error_msg)
                with ux_utils.print_exception_no_traceback():
                    raise RuntimeError(
                        ux_utils.error_message(error_msg + '\n',
                                               params.log_file,
                                               is_local=True))


@contextlib.contextmanager
def _setup_upload_logger(
        log_file: str) -> Generator[logging.Logger, None, None]:
    try:
        upload_logger = logging.getLogger('sky.upload')
        upload_logger.propagate = False
        handler = logging.FileHandler(os.path.expanduser(log_file),
                                      encoding='utf-8')
        handler.setFormatter(sky_logging.FORMATTER)
        upload_logger.addHandler(handler)
        upload_logger.setLevel(logging.DEBUG)
        yield upload_logger
    finally:
        upload_logger.removeHandler(handler)
        handler.close()


def upload_mounts_to_api_server(dag: 'sky.Dag',
                                workdir_only: bool = False) -> 'dag_lib.Dag':
    """Upload user files to remote API server.

    This function needs to be called after sdk.validate(),
    as the file paths need to be expanded to keep file_mounts_mapping
    aligned with the actual task uploaded to SkyPilot API server.

    We don't use FastAPI's built-in multipart upload, as nginx's
    client_max_body_size can block the request due to large request body, i.e.,
    even though the multipart upload streams the file to the server, there is
    only one HTTP request, and a large request body will be blocked by nginx.

    Args:
        dag: The dag where the file mounts are defined.
        workdir_only: Whether to only upload the workdir, which is used for
            `exec`, as it does not need other files/folders in file_mounts.

    Returns:
        The dag with the file_mounts_mapping updated, which maps the original
        file paths to the full path, so that on API server, the file paths can
        be retrieved by adding prefix to the full path.
    """

    if server_common.is_api_server_local():
        return dag

    def _full_path(src: str) -> str:
        return os.path.abspath(os.path.expanduser(src))

    upload_list = []
    for task_ in dag.tasks:
        task_.file_mounts_mapping = {}
        if task_.workdir and isinstance(task_.workdir, str):
            workdir = task_.workdir
            assert os.path.isabs(workdir)
            upload_list.append(workdir)
            task_.file_mounts_mapping[workdir] = workdir
        if workdir_only:
            continue
        if task_.file_mounts is not None:
            for src in task_.file_mounts.values():
                if not data_utils.is_cloud_store_url(src):
                    assert os.path.isabs(src)
                    upload_list.append(src)
                    task_.file_mounts_mapping[src] = src
                if src == constants.LOCAL_SKYPILOT_CONFIG_PATH_PLACEHOLDER:
                    # The placeholder for the local skypilot config path is in
                    # file mounts for controllers. It will be replaced with the
                    # real path for config file on API server.
                    pass
        if task_.storage_mounts is not None:
            for storage in task_.storage_mounts.values():
                storage_source = storage.source
                is_cloud_store_url = (
                    isinstance(storage_source, str) and
                    data_utils.is_cloud_store_url(storage_source))
                if (storage_source is not None and not is_cloud_store_url):
                    if isinstance(storage_source, str):
                        storage_source = [storage_source]
                    for src in storage_source:
                        upload_list.append(_full_path(src))
                        task_.file_mounts_mapping[src] = _full_path(src)
        if (task_.service is not None and
                task_.service.tls_credential is not None):
            keyfile = task_.service.tls_credential.keyfile
            certfile = task_.service.tls_credential.certfile
            upload_list.append(_full_path(keyfile))
            upload_list.append(_full_path(certfile))
            task_.file_mounts_mapping[keyfile] = _full_path(keyfile)
            task_.file_mounts_mapping[certfile] = _full_path(certfile)

    if upload_list:
        os.makedirs(os.path.expanduser(FILE_UPLOAD_LOGS_DIR), exist_ok=True)
        upload_id = sky_logging.get_run_timestamp()
        upload_id = f'{upload_id}-{uuid.uuid4().hex[:8]}'
        log_file = os.path.join(FILE_UPLOAD_LOGS_DIR, f'{upload_id}.log')

        logger.info(ux_utils.starting_message('Uploading files to API server'))
        with rich_utils.client_status(
                ux_utils.spinner_message(
                    'Uploading files to API server (1/2 - Zipping)',
                    log_file,
                    is_local=True)) as status, _setup_upload_logger(
                        log_file) as upload_logger:
            with tempfile.NamedTemporaryFile(suffix='.zip',
                                             delete=False) as temp_zip_file:
                upload_logger.info(
                    f'Zipping files to be uploaded: {upload_list}')
                storage_utils.zip_files_and_folders(upload_list,
                                                    temp_zip_file.name)
                upload_logger.info(f'Zipped files to: {temp_zip_file.name}')

            zip_file_size = os.path.getsize(temp_zip_file.name)
            # Per chunk size 512 MB
            total_chunks = int(math.ceil(zip_file_size / _UPLOAD_CHUNK_BYTES))
            timeout = httpx.Timeout(None, read=180.0)
            status.update(
                ux_utils.spinner_message(
                    'Uploading files to API server (2/2 - Uploading)',
                    log_file,
                    is_local=True))

            with httpx.Client(timeout=timeout) as client:
                chunk_params = [
                    UploadChunkParams(client, upload_id, chunk_index,
                                      total_chunks, temp_zip_file.name,
                                      upload_logger, log_file)
                    for chunk_index in range(total_chunks)
                ]
                subprocess_utils.run_in_parallel(_upload_chunk_with_retry,
                                                 chunk_params)
        os.unlink(temp_zip_file.name)
        upload_logger.info(f'Uploaded files: {upload_list}')
        logger.info(
            ux_utils.finishing_message('Files uploaded',
                                       log_file,
                                       is_local=True))

    return dag



"""ProcessPoolExecutor with additional supports for skypilot."""
import concurrent.futures
import logging
import multiprocessing
import threading
import time
from typing import Callable, Dict, Optional, Tuple

from sky import exceptions
from sky.utils import atomic
from sky.utils import subprocess_utils

logger = logging.getLogger(__name__)


class PoolExecutor(concurrent.futures.ProcessPoolExecutor):
    """A custom ProcessPoolExecutor with additional supports for skypilot.

    The additional supports include:
    1. Disposable workers: support control whether the worker process should
       exit after complete a task.
    2. Idle check: support check if there are any idle workers.
    3. Proactive shutdown: SIGTERM worker processes when the executor is
       shutting down instead of indefinitely waiting.
    """

    def __init__(self, max_workers: int, **kwargs):
        super().__init__(max_workers=max_workers, **kwargs)
        self.max_workers: int = max_workers
        # The number of workers that are handling tasks, atomicity across
        # multiple threads is sufficient since the idleness check is
        # best-effort and does not affect the correctness.
        # E.g. the following case is totally fine:
        # 1. Thread 1 checks running == max_workers
        # 2. Thread 2 decrements running
        # 3. Thread 1 schedules the task to other pool even if the pool is
        #    currently idle.
        self.running: atomic.AtomicInt = atomic.AtomicInt(0)

    def submit(self, fn, *args, **kwargs) -> concurrent.futures.Future:
        """Submit a task for execution.

        If reuse_worker is False, wraps the function to exit after completion.
        """
        self.running.increment()
        future = super().submit(fn, *args, **kwargs)
        future.add_done_callback(lambda _: self.running.decrement())
        return future

    def has_idle_workers(self) -> bool:
        """Check if there are any idle workers."""
        return self.running.get() < self.max_workers

    def shutdown(self, wait: bool = True) -> None:
        """Shutdown the executor."""
        # Here wait means wait for the proactive cancellation complete.
        # TODO(aylei): we may support wait=True in the future if needed.
        assert wait is True, 'wait=False is not supported'
        executor_processes = list(self._processes.values())
        # Shutdown the executor so that executor process can exit once the
        # running task is finished or interrupted.
        super().shutdown(wait=False)
        # Proactively interrupt the running task to avoid indefinite waiting.
        subprocess_utils.run_in_parallel(
            subprocess_utils.kill_process_with_grace_period,
            executor_processes,
            num_threads=len(executor_processes))


# Define the worker function outside of the class to avoid pickling self
def _disposable_worker(fn, initializer, initargs, result_queue, args, kwargs):
    """The worker function that is used to run the task.

    Args:
        fn: The function to run.
        initializer: The initializer function to run before running the task.
        initargs: The arguments to pass to the initializer function.
        result_queue: The queue to put the result and exception into.
        args: The arguments to pass to the function.
        kwargs: The keyword arguments to pass to the function.
    """
    try:
        if initializer is not None:
            initializer(*initargs)
        result = fn(*args, **kwargs)
        result_queue.put(result)
    except BaseException as e:  # pylint: disable=broad-except
        result_queue.put(e)


class DisposableExecutor:
    """A simple wrapper that creates a new process for each task.

    This is a workaround for Python 3.10 since `max_tasks_per_child` of
    ProcessPoolExecutor was introduced in 3.11. There is no way to control
    the worker lifetime in 3.10.
    Ref: https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor # pylint: disable=line-too-long
    TODO(aylei): use the official `max_tasks_per_child` when upgrade to 3.11
    """

    def __init__(self,
                 max_workers: Optional[int] = None,
                 initializer: Optional[Callable] = None,
                 initargs: Tuple = ()):
        self.max_workers: Optional[int] = max_workers
        self.workers: Dict[int, multiprocessing.Process] = {}
        self._shutdown: bool = False
        self._lock: threading.Lock = threading.Lock()
        self._initializer: Optional[Callable] = initializer
        self._initargs: Tuple = initargs

    def _monitor_worker(self, process: multiprocessing.Process,
                        future: concurrent.futures.Future,
                        result_queue: multiprocessing.Queue) -> None:
        """Monitor the worker process and cleanup when it's done."""
        try:
            process.join()
            if not future.cancelled():
                try:
                    # Get result from the queue if process completed
                    if not result_queue.empty():
                        result = result_queue.get(block=False)
                        if isinstance(result, BaseException):
                            future.set_exception(result)
                        else:
                            future.set_result(result)
                    else:
                        # Process ended but no result
                        future.set_result(None)
                except (multiprocessing.TimeoutError, BrokenPipeError,
                        EOFError) as e:
                    future.set_exception(e)
        finally:
            if process.pid:
                with self._lock:
                    if process.pid in self.workers:
                        del self.workers[process.pid]

    def submit(self, fn, *args, **kwargs) -> concurrent.futures.Future:
        """Submit a task for execution and return a Future."""
        future: concurrent.futures.Future = concurrent.futures.Future()

        if self._shutdown:
            raise RuntimeError('Cannot submit task after executor is shutdown')

        with self._lock:
            if (self.max_workers is not None and
                    len(self.workers) >= self.max_workers):
                raise exceptions.ExecutionPoolFullError(
                    'Maximum workers reached')

        result_queue: multiprocessing.Queue = multiprocessing.Queue()
        process = multiprocessing.Process(target=_disposable_worker,
                                          args=(fn, self._initializer,
                                                self._initargs, result_queue,
                                                args, kwargs))
        process.daemon = True
        process.start()

        with self._lock:
            pid = process.pid or 0
            if pid == 0:
                raise RuntimeError('Failed to start process')
            self.workers[pid] = process

        # Start monitor thread to cleanup the worker process when it's done
        monitor_thread = threading.Thread(target=self._monitor_worker,
                                          args=(process, future, result_queue),
                                          daemon=True)
        monitor_thread.start()

        return future

    def has_idle_workers(self) -> bool:
        """Check if there are any idle workers."""
        if self.max_workers is None:
            return True
        with self._lock:
            return len(self.workers) < self.max_workers

    def shutdown(self):
        """Shutdown the executor."""
        with self._lock:
            self._shutdown = True
        subprocess_utils.run_in_parallel(
            subprocess_utils.kill_process_with_grace_period,
            list(self.workers.values()),  # Convert dict values to list
            num_threads=len(self.workers))


class BurstableExecutor:
    """An multiprocessing executor that supports bursting worker processes."""

    # _executor is a PoolExecutor that is used to run guaranteed requests.
    _executor: Optional[PoolExecutor] = None
    # _burst_executor is a ProcessPoolExecutor that is used to run burst
    # requests.
    _burst_executor: Optional[DisposableExecutor] = None

    def __init__(self,
                 garanteed_workers: int,
                 burst_workers: int = 0,
                 **kwargs):
        if garanteed_workers > 0:
            self._guaranteed_workers = garanteed_workers
            self._guaranteed_pool_kwargs = kwargs
            self._guaranteed_pool_lock = threading.Lock()
            self._executor = PoolExecutor(max_workers=garanteed_workers,
                                          **kwargs)
        if burst_workers > 0:
            self._burst_executor = DisposableExecutor(max_workers=burst_workers,
                                                      **kwargs)

    def submit_until_success(self, fn, *args,
                             **kwargs) -> concurrent.futures.Future:
        """Submit a task for execution until success.

        Prioritizes submitting to the guaranteed pool. If no idle workers
        are available in the guaranteed pool, it will submit to the burst
        pool. If the burst pool is full, it will retry the whole process until
        the task is submitted successfully.
        TODO(aylei): this is coupled with executor.RequestWorker since we
        know the worker is dedicated to request scheduling and it either
        blocks on request polling or request submitting. So it is no harm
        to make submit blocking here. But for general cases, we need an
        internal queue to decouple submit and run.
        """

        while True:
            if self._executor is not None and self._executor.has_idle_workers():
                return self._submit_to_guaranteed_pool(fn, *args, **kwargs)
            if (self._burst_executor is not None and
                    self._burst_executor.has_idle_workers()):
                try:
                    fut = self._burst_executor.submit(fn, *args, **kwargs)
                    return fut
                except exceptions.ExecutionPoolFullError:
                    # The burst pool is full, try the next candidate.
                    pass
            if self._executor is not None:
                # No idle workers in either pool, still queue the request
                # to the guaranteed pool to keep behavior consistent.
                return self._submit_to_guaranteed_pool(fn, *args, **kwargs)
            logger.debug('No guaranteed pool set and the burst pool is full, '
                         'retry later.')
            time.sleep(0.1)

    def _submit_to_guaranteed_pool(self, fn, *args,
                                   **kwargs) -> concurrent.futures.Future:
        """Submit to the guaranteed pool with retry.

        The concurrent.futures.ProcessPoolExecutor will terminate all child
        processes and reject new tasks if any of the child process died, e.g.
        OOM killed. So when the pool is broken, we have to rebuild a new pool
        to execute tasks.

        Alternatives considered:
        - multiprocessing.Pool: individual process failure does not affect the
            pool, but it lacks of lazy-init support and does not handle process
            failure gracefully.
        - inherit from concurrent.futures.ProcessPoolExecutor: If a child
            process dies when executing a task, we have to set an exception
            in the Future object of the task. But there is a race condition
            that the child process may dies exactly after get the task from
            queue so there is no (obviously) safe way to record process PID
            to task mapping in the current architecture of
            ProcessPoolExecutor. It is more feasible to implement a custom
            ProcessPoolExecutor with an new architecture.
        - a custom ProcessPoolExecutor: non-trivial to implement, keep for
            future improvement as BrokenProcessPool is not expected to be
            a common case, i.e. it usually indicates there a bug elsewhere
            when this happens.
        """
        with self._guaranteed_pool_lock:
            assert self._executor is not None
            try:
                return self._executor.submit(fn, *args, **kwargs)
            except concurrent.futures.process.BrokenProcessPool as e:
                logger.warning('The guaranteed pool is broken, '
                               f'replacing the pool and retrying. Error: {e}')
                broken_pool = self._executor
                threading.Thread(target=broken_pool.shutdown,
                                 daemon=True).start()
                self._executor = PoolExecutor(
                    max_workers=self._guaranteed_workers,
                    **self._guaranteed_pool_kwargs)
        return self._submit_to_guaranteed_pool(fn, *args, **kwargs)

    def shutdown(self) -> None:
        """Shutdown the executor."""

        if self._burst_executor is not None:
            self._burst_executor.shutdown()
        if self._executor is not None:
            self._executor.shutdown(wait=True)



"""Constants used in the SkyPilot provisioner."""

# Tag uniquely identifying all nodes of a cluster
TAG_RAY_CLUSTER_NAME = 'ray-cluster-name'
TAG_SKYPILOT_CLUSTER_NAME = 'skypilot-cluster-name'
# Legacy tag for backward compatibility to distinguish head and worker nodes.
TAG_RAY_NODE_KIND = 'ray-node-type'
TAG_SKYPILOT_HEAD_NODE = 'skypilot-head-node'

HEAD_NODE_TAGS = {
    TAG_RAY_NODE_KIND: 'head',
    TAG_SKYPILOT_HEAD_NODE: '1',
}

WORKER_NODE_TAGS = {
    TAG_RAY_NODE_KIND: 'worker',
    TAG_SKYPILOT_HEAD_NODE: '0',
}

# Names for Azure Deployments.
DEPLOYMENT_NAME = 'skypilot-config'
LEGACY_DEPLOYMENT_NAME = 'ray-config'
EXTERNAL_RG_BOOTSTRAP_DEPLOYMENT_NAME = (
    'skypilot-bootstrap-{cluster_name_on_cloud}')
EXTERNAL_RG_VM_DEPLOYMENT_NAME = 'skypilot-vm-{cluster_name_on_cloud}'



"""Vsphere custom script."""
CUSTOMIZED_SCRIPT = """#!/bin/bash
if [ x$1 = x"precustomization" ]; then
    sudo mkdir -p /home/user_placeholder/.ssh
    sudo echo "ssh_public_key" | sudo tee /home/user_placeholder/.ssh/authorized_keys
    sudo mkdir -p /home/user_placeholder/.ssh/pre
    sudo chown -R user_placeholder:user_placeholder /home/user_placeholder/.ssh
    sudo chmod 700 /home/user_placeholder/.ssh
    sudo chmod 644 /home/user_placeholder/.ssh/authorized_keys
fi"""



"""Responses for the API server."""

from typing import Any, Dict, List, Optional

import pydantic

from sky import models
from sky.server import common
from sky.utils import status_lib


class ResponseBaseModel(pydantic.BaseModel):
    """A pydantic model that acts like a dict.

    Supports the following syntax:
    class SampleResponse(DictLikePayload):
        field: str

    response = SampleResponse(field='value')
    print(response['field']) # prints 'value'
    response['field'] = 'value2'
    print(response['field']) # prints 'value2'
    print('field' in response) # prints True

    This model exists for backwards compatibility with the
    old SDK that used to return a dict.

    The backward compatibility may be removed
    in the future.
    """
    # Ignore extra fields in the request body, which is useful for backward
    # compatibility. The difference with `allow` is that `ignore` will not
    # include the unknown fields when dump the model, i.e., we can add new
    # fields to the request body without breaking the existing old API server
    # where the handler function does not accept the new field in function
    # signature.
    model_config = pydantic.ConfigDict(extra='ignore')

    # backward compatibility with dict
    # TODO(syang): remove this in v0.13.0
    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError as e:
            raise KeyError(key) from e

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __contains__(self, key):
        return hasattr(self, key)

    def keys(self):
        return self.model_dump().keys()

    def values(self):
        return self.model_dump().values()

    def items(self):
        return self.model_dump().items()

    def __repr__(self):
        return self.__dict__.__repr__()


class APIHealthResponse(ResponseBaseModel):
    """Response for the API health endpoint."""
    status: common.ApiServerStatus
    api_version: str = ''
    version: str = ''
    version_on_disk: str = ''
    commit: str = ''
    basic_auth_enabled: bool = False
    user: Optional[models.User] = None


class StatusResponse(ResponseBaseModel):
    """Response for the status endpoint."""
    name: str
    launched_at: int
    # pydantic cannot generate the pydantic-core schema for
    # backends.ResourceHandle, so we use Any here.
    # This is an internally facing field anyway, so it's less
    # of a problem that it's not typed.
    handle: Any
    last_use: str
    status: status_lib.ClusterStatus
    autostop: int
    to_down: bool
    owner: Optional[List[str]] = None
    # metadata is a JSON, so we use Any here.
    metadata: Optional[Dict[str, Any]] = None
    cluster_hash: str
    # pydantic cannot generate the pydantic-core schema for
    # storage_mounts_metadata, so we use Any here.
    storage_mounts_metadata: Optional[Dict[str, Any]] = None
    cluster_ever_up: bool
    status_updated_at: int
    user_hash: str
    user_name: str
    config_hash: Optional[str] = None
    workspace: str
    last_creation_yaml: Optional[str] = None
    last_creation_command: Optional[str] = None
    is_managed: bool
    last_event: Optional[str] = None
    resources_str: Optional[str] = None
    resources_str_full: Optional[str] = None
    # credentials is a JSON, so we use Any here.
    credentials: Optional[Dict[str, Any]] = None
    nodes: int
    cloud: Optional[str] = None
    region: Optional[str] = None
    cpus: Optional[str] = None
    memory: Optional[str] = None
    accelerators: Optional[str] = None
